import posixpath
from datetime import datetime
from munch import Munch
from twisted.internet import defer
from txkoji import build_states


class Build(Munch):

    @property
    def completed(self):
        """
        Return a parsed completion datetime for a build.

        :returns: a datetime object for the time this build finished,
                  or None if the build is not completed.
        """
        if not self.completion_ts:
            return None
        return datetime.utcfromtimestamp(self.completion_ts)

    @property
    def started(self):
        """
        Return a parsed started datetime for a build.

        :returns: a datetime object for the time this build started
        """
        return datetime.utcfromtimestamp(self.start_ts)

    @property
    def duration(self):
        """
        Return a timedelta for this build.

        Measure the time between this build's start and end time, or "now"
        if the build has not yet finished.

        :returns: timedelta object
        """
        if self.completion_ts:
            end = self.completed
        else:
            end = datetime.utcnow()
        return end - self.started

    @property
    def url(self):
        """
        Return a kojiweb URL for this resource.

        :returns: ``str``, kojiweb URL like
                  "http://cbs.centos.org/koji/buildinfo?buildID=21155"
        """
        endpoint = 'buildinfo?buildID=%d' % self.build_id
        return posixpath.join(self.connection.weburl, endpoint)

    @defer.inlineCallbacks
    def estimate_completion(self):
        """
        Estimate completion time for a build.

        This calls getAverageBuildDuration on the hub for this package. This
        value is a very rough guess, an average for all completed builds in the
        system.

        For now this is better than nothing, but I'm recording a few thoughts
        here for posterity:

        A more advanced implementation of for estimating build times would
        track unique averages per build target. Early releases of Ceph "Hammer"
        versions would build much more quickly than newer Ceph releases like
        "Mimic", and the current method simply averages all of them.

        Also, different Koji build targets can have different arches. When we
        build ceph in "newarch" side tags, those build times are wildly beyond
        the average, but they influence this average too, subtly throwing it
        off for the x86-only builds that I care about.

        :returns: deferred that when fired returns a datetime object for the
                  estimated or actual datetime.
        """
        if self.state != build_states.BUILDING:
            # Build is already complete. Return the exact completion time:
            defer.returnValue(self.completed)
        avg_delta = yield self.connection.getAverageBuildDuration(self.name)
        est_completion = self.started + avg_delta
        defer.returnValue(est_completion)



from twisted.python.compat import long
try:
    import xmlrpc
    from xmlrpc.client import Marshaller
except ImportError:
    import xmlrpclib as xmlrpc
    from xmlrpclib import Marshaller


class KojiMarshaller(Marshaller):
    """ Custom XML-RPC Marshaller for long ints. """
    dispatch = Marshaller.dispatch.copy()

    MAXI8 = 2 ** 63 - 1
    MINI8 = -2 ** 63

    def dump_int(self, value, write):
        if (value > self.MAXI8 or value < self.MINI8):
            raise OverflowError("long int exceeds XML-RPC limits")
        elif value > xmlrpc.MAXINT or value < xmlrpc.MININT:
            write("<!-- using i8 extension -->")
            write("<value><i8>")
            write(str(int(value)))
            write("</i8></value>\n")
        else:
            return Marshaller.dump_int(self, value, write)

    dispatch[int] = dump_int
    dispatch[long] = dump_int



import posixpath
from munch import Munch


class Package(Munch):

    @property
    def url(self):
        """
        Return a kojiweb URL for this resource.

        :returns: ``str``, kojiweb URL like
                  "http://cbs.centos.org/koji/packageinfo?packageID=3726"
        """
        endpoint = 'packageinfo?packageID=%d' % self.id
        return posixpath.join(self.connection.weburl, endpoint)



from datetime import timedelta
from glob import glob
import os
from munch import munchify
import treq_kerberos
from twisted.web.xmlrpc import Proxy
from twisted.internet import defer
from twisted.web.client import ResponseFailed
try:
    from configparser import SafeConfigParser
    from urllib.parse import urlencode, urlparse, parse_qs
    import xmlrpc
except ImportError:
    from ConfigParser import SafeConfigParser
    from urllib import urlencode
    from urlparse import urlparse, parse_qs
    import xmlrpclib as xmlrpc
from txkoji.query_factory import KojiQueryFactory
from txkoji.task import Task
from txkoji.build import Build
from txkoji.package import Package


__version__ = '0.3.0'


PROFILES = '/etc/koji.conf.d/*.conf'


class Call(object):
    """
    Callable abstract class representing a Koji RPC, eg "getTag".

    :param connection: ``txkoji.Connection``
    :param name: XML-RPC method name to call on the server, eg. "getTag".
    """
    def __init__(self, connection, name):
        self.connection = connection
        self.name = name

    def __call__(self, *args, **kwargs):
        return self.connection.call(self.name, *args, **kwargs)


class Connection(object):

    def __init__(self, profile):
        self.profile = profile
        self.url = self.lookup(profile, 'server')
        self.weburl = self.lookup(profile, 'weburl')
        if not self.url:
            msg = 'no server configured at %s for %s' % (PROFILES, profile)
            raise ValueError(msg)
        self.proxy = Proxy(self.url.encode(), allowNone=True)
        self.proxy.queryFactory = KojiQueryFactory
        # We populate these on login:
        self.session_id = None
        self.session_key = None
        self.callnum = None

    def lookup(self, profile, setting):
        """ Check koji.conf.d files for this profile's setting.

        :param setting: ``str`` like "server" (for kojihub) or "weburl"
        :returns: ``str``, value for this setting
        """
        for path in glob(PROFILES):
            cfg = SafeConfigParser()
            cfg.read(path)
            if profile not in cfg.sections():
                continue
            if not cfg.has_option(profile, setting):
                continue
            return cfg.get(profile, setting)

    @classmethod
    def connect_from_web(klass, url):
        """
        Find a connection that matches this kojiweb URL.

        Check all koji.conf.d files' kojiweb URLs and load the profile that
        matches the url we pass in here.

        For example, if a user pastes a kojiweb URL into chat, we want to
        discover the corresponding Koji instance hub automatically.

        :param url: ``str``, for example
                    "http://cbs.centos.org/koji/buildinfo?buildID=21155"
        :returns: A "Connection" instance
        """
        for path in glob(PROFILES):
            cfg = SafeConfigParser()
            cfg.read(path)
            for profile in cfg.sections():
                if not cfg.has_option(profile, 'weburl'):
                    continue
                weburl = cfg.get(profile, 'weburl')
                if weburl in url:
                    return klass(profile)

    @defer.inlineCallbacks
    def from_web(self, url):
        """
        Reverse-engineer a kojiweb URL into an equivalent API response.

        Only a few kojiweb URL endpoints work here.

        See also connect_from_web().

        :param url: ``str``, for example
                    "http://cbs.centos.org/koji/buildinfo?buildID=21155"
        :returns: deferred that when fired returns a Munch (dict-like) object
                  with data about this resource, or None if we could not parse
                  the url.
        """
        o = urlparse(url)
        endpoint = os.path.basename(o.path)
        if o.query:
            query = parse_qs(o.query)
        result = None
        # Known Kojiweb endpoints:
        endpoints = {
            'buildinfo': ('buildID', self.getBuild),
            'packageinfo': ('packageID', self.getPackage),
            'taskinfo': ('taskID', self.getTaskInfo),
            'taginfo': ('tagID', self.getTag),
            'targetinfo': ('targetID', self.getTarget),
            'userinfo': ('userID', self.getUser),
        }
        try:
            (param, method) = endpoints[endpoint]
            id_ = int(query[param][0])
            result = yield method(id_)
        except KeyError:
            pass
        defer.returnValue(result)

    def call(self, method, *args, **kwargs):
        """
        Make an XML-RPC call to the server.

        If this client is logged in (with login()), this call will be
        authenticated.

        Koji has its own custom implementation of XML-RPC that supports named
        args (kwargs). For example, to use the "queryOpts" kwarg:

          d = client.call('listBuilds', package_id,
                          queryOpts={'order': 'creation_ts'})

        In this example, the server will order the list of builds according to
        the "creation_ts" database column. Many of Koji's XML-RPC methods have
        optional kwargs that you can set as needed.

        :returns: deferred that when fired returns a dict with data from this
                  XML-RPC call.
        """
        if kwargs:
            kwargs['__starstar'] = True
            args = args + (kwargs,)
        if self.session_id:
            self.proxy.path = self._authenticated_path()
        d = self.proxy.callRemote(method, *args)
        d.addCallback(self._munchify_callback)
        d.addErrback(self._parse_errback)
        if self.callnum is not None:
            self.callnum += 1
        return d

    def _authenticated_path(self):
        """
        Get the path of our XML-RPC endpoint with session auth params added.

        For example:
          /kojihub?session-id=123456&session-key=1234-asdf&callnum=0

        If we're making an authenticated request, we must add these session
        parameters to the hub's XML-RPC endpoint.

        :return: a path suitable for twisted.web.xmlrpc.Proxy
        """
        basepath = self.proxy.path.decode().split('?')[0]
        params = urlencode({'session-id': self.session_id,
                            'session-key': self.session_key,
                            'callnum': self.callnum})
        result = '%s?%s' % (basepath, params)
        return result.encode('utf-8')

    def __getattr__(self, name):
        return Call(self, name)

    @defer.inlineCallbacks
    def getAverageBuildDuration(self, package, **kwargs):
        """
        Return a timedelta that Koji considers to be average for this package.

        Calls "getAverageBuildDuration" XML-RPC.

        :param package: ``str``, for example "ceph"
        :returns: deferred that when fired returns a datetime object for the
                  estimated duration.
        """
        seconds = yield self.call('getAverageBuildDuration', package, **kwargs)
        defer.returnValue(timedelta(seconds=seconds))

    @defer.inlineCallbacks
    def getBuild(self, build_id, **kwargs):
        """
        Load all information about a build and return a custom Build class.

        Calls "getBuild" XML-RPC.

        :param build_id: ``int``, for example 12345
        :returns: deferred that when fired returns a Build (Munch, dict-like)
                  object representing this Koji build, or None if no build was
                  found.
        """
        buildinfo = yield self.call('getBuild', build_id, **kwargs)
        build = Build.fromDict(buildinfo)
        if build:
            build.connection = self
        defer.returnValue(build)

    @defer.inlineCallbacks
    def getPackage(self, name, **kwargs):
        """
        Load information about a package and return a custom Package class.

        Calls "getPackage" XML-RPC.

        :param package_id: ``int``, for example 12345
        :returns: deferred that when fired returns a Package (Munch, dict-like)
                  object representing this Koji package, or None if no build
                  was found.
        """
        packageinfo = yield self.call('getPackage', name, **kwargs)
        package = Package.fromDict(packageinfo)
        if package:
            package.connection = self
        defer.returnValue(package)

    @defer.inlineCallbacks
    def getTaskDescendents(self, task_id, **kwargs):
        """
        Load all information about a task's descendents into Task classes.

        Calls "getTaskDescendents" XML-RPC (with request=True to get the full
        information.)

        :param task_id: ``int``, for example 12345, parent task ID
        :returns: deferred that when fired returns a list of Task (Munch,
                  dict-like) objects representing Koji tasks.
        """
        kwargs['request'] = True
        data = yield self.call('getTaskDescendents', task_id, **kwargs)
        tasks = []
        for tdata in data[str(task_id)]:
            task = Task.fromDict(tdata)
            task.connection = self
            tasks.append(task)
        defer.returnValue(tasks)

    @defer.inlineCallbacks
    def getTaskInfo(self, task_id, **kwargs):
        """
        Load all information about a task and return a custom Task class.

        Calls "getTaskInfo" XML-RPC (with request=True to get the full
        information.)

        :param task_id: ``int``, for example 12345
        :returns: deferred that when fired returns a Task (Munch, dict-like)
                  object representing this Koji task, or none if no task was
                  found.
        """
        kwargs['request'] = True
        taskinfo = yield self.call('getTaskInfo', task_id, **kwargs)
        task = Task.fromDict(taskinfo)
        if task:
            task.connection = self
        defer.returnValue(task)

    @defer.inlineCallbacks
    def listBuilds(self, package, **kwargs):
        """
        Get information about all builds of a package.

        Calls "listBuilds" XML-RPC, with an enhancement: you can also pass a
        string here for the package name instead of the package ID number.

        :param package: ``int`` (packageID) or ``str`` (package name).
        :returns: deferred that when fired returns a list of Build objects
                  for this package.
        """
        if isinstance(package, int):
            package_id = package
        else:
            package_data = yield self.getPackage(package)
            if package_data is None:
                defer.returnValue([])
            package_id = package_data.id
        data = yield self.call('listBuilds', package_id, **kwargs)
        builds = []
        for bdata in data:
            build = Build.fromDict(bdata)
            build.connection = self
            builds.append(build)
        defer.returnValue(builds)

    @defer.inlineCallbacks
    def listTasks(self, opts={}, queryOpts={}):
        """
        Get information about all Koji tasks.

        Calls "listTasks" XML-RPC.

        :param opts: ``dict``. Eg. {'state': [task_states.OPEN]}
        :param queryOpts: ``dict``. Eg. {'order' : 'priority,create_time'}
        :returns: deferred that when fired returns a list of Task objects.
        """
        opts['decode'] = True  # decode xmlrpc data in "request"
        data = yield self.call('listTasks', opts, queryOpts)
        tasks = []
        for tdata in data:
            task = Task.fromDict(tdata)
            task.connection = self
            tasks.append(task)
        defer.returnValue(tasks)

    @defer.inlineCallbacks
    def login(self):
        """
        Return True if we successfully logged into this Koji hub.

        We only support GSSAPI auth for now.

        :returns: deferred that when fired returns True
        """
        authtype = self.lookup(self.profile, 'authtype')
        if authtype is None:
            cert = self.lookup(self.profile, 'cert')
            if cert and os.path.isfile(os.path.expanduser(cert)):
                authtype = 'ssl'
            # Note: official koji cli is a little more lax here. If authtype is
            # None and we have a valid kerberos ccache, we still try kerberos
            # auth.
        if authtype == 'kerberos':
            # Note: we don't try the old-style kerberos login here.
            result = yield self._gssapi_login()
        else:
            raise NotImplementedError('non-kerberos auth: %s' % authtype)
        self.session_id = result['session-id']
        self.session_key = result['session-key']
        self.callnum = 0  # increment this on every call for this session.
        defer.returnValue(True)

    @defer.inlineCallbacks
    def _gssapi_login(self):
        """
        Authenticate to the /ssllogin endpoint with GSSAPI authentication.

        Raises KojiGssapiException if this fails.

        :returns: deferred that when fired returns a dict from sslLogin
        """
        auth = treq_kerberos.TreqKerberosAuth(force_preemptive=True)
        # Hit the special "/ssllogin" URL that's behind mod_auth_kerb:
        url = self.url + '/ssllogin'
        # Build the XML-RPC HTTP request body by hand and send it with
        # treq-kerberos.
        factory = KojiQueryFactory(path=None, host=None, method='sslLogin')
        payload = factory.payload
        try:
            response = yield treq_kerberos.post(url, data=payload, auth=auth)
        except ResponseFailed as e:
            failure = e.reasons[0]
            failure.raiseException()
            # raise KojiGssapiException(failure.value)
        if response.code > 200:
            raise KojiGssapiException('HTTP %d error' % response.code)
        # Process the XML-RPC response content from treq.
        content = yield response.content()
        if hasattr(xmlrpc, 'loads'):  # Python 2:
            result = xmlrpc.loads(content)[0][0]
        else:
            result = xmlrpc.client.loads(content)[0][0]
        defer.returnValue(result)

    def _munchify_callback(self, value):
        """
        Fires when we get user information back from the XML-RPC server.

        This is a generic callback for when we do not want to post-process the
        XML-RPC server's data further.

        :param value: dict of data from XML-RPC server.
        :returns: ``Munch`` (dict-like) object
        """
        return munchify(value)

    def _parse_errback(self, error):
        """
        Parse an error from an XML-RPC call.

        raises: ``IOError`` when the Twisted XML-RPC connection times out.
        raises: ``KojiException`` if we got a response from the XML-RPC
                server but it is not one of the ``xmlrpc.Fault``s that
                we know about.
        raises: ``Exception`` if it is not one of the above.
        """
        if isinstance(error.value, IOError):
            raise error.value
        if hasattr(xmlrpc, 'Fault'):  # Python 2:
            fault = xmlrpc.Fault
        else:
            fault = xmlrpc.client.Fault
        if isinstance(error.value, fault):
            # TODO: specific errors here, see koji/__init__.py
            if error.value.faultCode >= 1000 and error.value.faultCode <= 1022:
                raise KojiException(error.value.faultString)
            raise KojiException(error.value)
        # We don't know what this is, so just raise it.
        raise error


class KojiException(Exception):
    pass


class KojiGssapiException(Exception):
    pass



import os
import re
import subprocess
import sys
from setuptools.command.test import test as TestCommand
from setuptools import setup, Command
try:
    # Python 2 backwards compat
    from __builtin__ import raw_input as input
except ImportError:
    pass

readme = os.path.join(os.path.dirname(__file__), 'README.rst')
LONG_DESCRIPTION = open(readme).read()


def read_module_contents():
    with open('txkoji/__init__.py') as txkoji_init:
        return txkoji_init.read()


module_file = read_module_contents()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))
version = metadata['version']


class BumpCommand(Command):
    """ Bump the __version__ number and commit all changes. """

    user_options = [('version=', 'v', 'version number to use')]

    def initialize_options(self):
        new_version = metadata['version'].split('.')
        new_version[-1] = str(int(new_version[-1]) + 1)  # Bump the final part
        self.version = ".".join(new_version)

    def finalize_options(self):
        pass

    def run(self):

        print('old version: %s  new version: %s' %
              (metadata['version'], self.version))
        try:
            input('Press enter to confirm, or ctrl-c to exit >')
        except KeyboardInterrupt:
            raise SystemExit("\nNot proceeding")

        old = "__version__ = '%s'" % metadata['version']
        new = "__version__ = '%s'" % self.version
        module_file = read_module_contents()
        with open('txkoji/__init__.py', 'w') as fileh:
            fileh.write(module_file.replace(old, new))

        old = 'Version:        %s' % metadata['version']
        new = 'Version:        %s' % self.version

        # Commit everything with a standard commit message
        cmd = ['git', 'commit', '-a', '-m', 'version %s' % self.version]
        print(' '.join(cmd))
        subprocess.check_call(cmd)


class ReleaseCommand(Command):
    """ Tag and push a new release. """

    user_options = [('sign', 's', 'GPG-sign the Git tag and release files')]

    def initialize_options(self):
        self.sign = False

    def finalize_options(self):
        pass

    def run(self):
        # Create Git tag
        tag_name = 'v%s' % version
        cmd = ['git', 'tag', '-a', tag_name, '-m', 'version %s' % version]
        if self.sign:
            cmd.append('-s')
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push Git tag to origin remote
        cmd = ['git', 'push', 'origin', tag_name]
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push master branch to origin remote
        cmd = ['git', 'push', 'origin', 'master']
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Create source package
        cmd = ['python', 'setup.py', 'sdist']
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        tarball = 'dist/%s-%s.tar.gz' % ('txkoji', version)

        # GPG sign
        if self.sign:
            cmd = ['gpg2', '-b', '-a', tarball]
            print(' '.join(cmd))
            subprocess.check_call(cmd)

        # Upload
        cmd = ['twine', 'upload', tarball]
        if self.sign:
            cmd.append(tarball + '.asc')
        print(' '.join(cmd))
        subprocess.check_call(cmd)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main('txkoji --flake8 ' +
                            self.pytest_args)
        sys.exit(errno)


setup(
    name='txkoji',
    description='Twisted API for Koji',
    packages=['txkoji'],
    author='Ken Dreyer',
    author_email='ktdreyer@ktdreyer.com',
    url='https://github.com/ktdreyer/txkoji',
    version=metadata['version'],
    license='MIT',
    zip_safe=False,
    keywords='koji, twisted',
    long_description=LONG_DESCRIPTION,
    classifiers=[
                 'Development Status :: 4 - Beta',
                 'Framework :: Twisted',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'munch',
        'pyOpenSSL',  # for Kojis that use HTTPS
        'service_identity',  # for Kojis that use HTTPS
        'pyasn1',  # for Kojis that use HTTPS
        'treq-kerberos',
        'twisted',
    ],
    tests_require=[
        'pytest-flake8',
        'pytest-twisted',
    ],
    cmdclass={'test': PyTest, 'bump': BumpCommand, 'release': ReleaseCommand},
)



from twisted.web.xmlrpc import _QueryFactory, payloadTemplate
from twisted.python.compat import unicode
from twisted.internet import defer
from txkoji.marshaller import KojiMarshaller


class KojiQueryFactory(_QueryFactory):
    def __init__(self, path, host, method, user=None, password=None,
                 allowNone=True, args=(), canceller=None, useDateTime=False):
        """
        Mainly copied from Twisted's _QueryFactory class, except we use our own
        KojiMarshaller instead of stdlib's xmlrpc Marshaller.
        """
        self.path, self.host = path, host
        self.user, self.password = user, password
        self.marshaller = KojiMarshaller('utf-8', allow_none=True)
        self.payload = payloadTemplate \
            % (method, self.marshaller.dumps(args))
        if isinstance(self.payload, unicode):
            self.payload = self.payload.encode('utf-8')
        self.deferred = defer.Deferred(canceller)
        self.useDateTime = useDateTime
        # Debug the client's XML-RPC payload:
        # print(self.payload)

    # def parseResponse(self, contents):
    #     """ Debugging: print the server's response to STDOUT. """
    #     print(contents)
    #     return _QueryFactory.parseResponse(self, contents)



from datetime import datetime
import os.path
import posixpath
from munch import Munch
from twisted.internet import defer
from txkoji import task_states
try:
    from urllib.parse import urlparse
    import xmlrpc
except ImportError:
    from urlparse import urlparse
    import xmlrpclib as xmlrpc


class Task(Munch):

    @property
    def completed(self):
        """
        Return a parsed completion datetime for a task.

        :returns: a datetime object for the time this task finished,
                  or None if the task is not completed.
        """
        if not self.completion_ts:
            return None
        return datetime.utcfromtimestamp(self.completion_ts)

    @property
    def started(self):
        """
        Return a parsed started datetime for a task.

        :returns: a datetime object for the time this task started
        """
        if not self.start_ts:
            return None
        return datetime.utcfromtimestamp(self.start_ts)

    @property
    def duration(self):
        """
        Return a timedelta for this task.

        Measure the time between this task's start and end time, or "now"
        if the task has not yet finished.

        :returns: timedelta object, or None if the task has not even started.
        """
        if not self.started:
            return None
        start = self.started
        end = self.completed
        if not end:
            end = datetime.utcnow()
        return end - start

    @defer.inlineCallbacks
    def estimate_completion(self):
        """
        Estimate completion time for a task.

        :returns: deferred that when fired returns a datetime object for the
                  estimated or actual datetime.
        """
        if self.completion_ts:
            # Task is already complete. Return the exact completion time:
            defer.returnValue(self.completed)
        # Get the timestamps from the descendent task that's doing the work:
        if self.method == 'build' or self.method == 'image':
            subtask_completion = yield self.estimate_descendents()
            defer.returnValue(subtask_completion)
        if not self.start_ts:
            raise ValueError('no start time, task in %s state' % self.state)
        package = self.package
        avg_delta = yield self.connection.getAverageBuildDuration(package)
        est_completion = self.started + avg_delta
        defer.returnValue(est_completion)

    @defer.inlineCallbacks
    def estimate_descendents(self):
        """
        Estimate from the descendent (child) tasks.

        :returns: deferred that when fired returns a datetime object for the
                  estimated or actual datetime.
        """
        if self.method == 'build':
            child_method = 'buildArch'
        if self.method == 'image':
            child_method = 'createImage'
        # Find the open child task and estimate that.
        subtasks = yield self.descendents(method=child_method,
                                          state=task_states.OPEN)
        if not subtasks:
            # Maybe koji has not assigned this task to a worker (over
            # capacity), or maybe makeSRPMfromSCM is still running.
            # Try again in a few minutes?
            raise ValueError('no running %s for task %d' %
                             (child_method, self.id))
        # Find subtask with the most recent start time:
        build_task = subtasks[0]
        for subtask in subtasks:
            if subtask.start_ts > build_task.start_ts:
                build_task = subtask
        subtask_completion = yield build_task.estimate_completion()
        defer.returnValue(subtask_completion)

    @defer.inlineCallbacks
    def descendents(self, method=None, state=None):
        """
        Find descendant tasks, optionally filtered by method and/or state.

        :param method: (optional) filter for tasks, eg. "buildArch".
        :param state: (optional) filter for tasks, eg. task_states.OPEN.
        :returns: deferred that when fired returns a list of Tasks.
        """
        subtasks = yield self.connection.getTaskDescendents(self.id)
        if method:
            subtasks = [t for t in subtasks if t.method == method]
        if state:
            subtasks = [t for t in subtasks if t.state == state]
        defer.returnValue(subtasks)

    @property
    def package(self):
        """
        Find a package name from a build task's parameters.

        :returns: name of the package this build task is building.
        :raises: ValueError if we could not parse this tasks's request params.
        """
        # (I wish there was a better way to do this.)
        source = self.params[0]
        o = urlparse(source)
        # build tasks can load an SRPM from a "cli-build" tmpdir:
        if source.endswith('.src.rpm'):
            srpm = os.path.basename(source)
            (name, version, release) = srpm.rsplit('-', 2)
            # Note we're throwing away version and release here. They could be
            # useful eventually, maybe in a "Package" class.
            return name
        # or an allowed SCM:
        elif o.scheme:
            package = os.path.basename(o.path)
            if package.endswith('.git'):
                return package[:-4]
            return package
        raise ValueError('could not parse source "%s"' % source)

    @property
    def params(self):
        """
        Return a list of parameters in this task's request.

        If self.request is already a list, simply return it.

        If self.request is a raw XML-RPC string, parse it and return the
        params.
        """
        if isinstance(self.request, list):
            return self.request
        (params, _) = xmlrpc.loads(self.request)
        return params

    @property
    def is_scratch(self):
        for param in self.params:
            if isinstance(param, dict):
                if param.get('scratch'):
                    return True
        return False

    @property
    def url(self):
        """
        Return a kojiweb URL for this resource.

        :returns: ``str``, kojiweb URL like
                  "http://cbs.centos.org/koji/taskinfo?taskID=381617"
        """
        endpoint = 'taskinfo?taskID=%d' % self.id
        return posixpath.join(self.connection.weburl, endpoint)
